from probatus.utils import assure_pandas_df, shap_calc, assure_list_of_strings, calculate_shap_importance, \
    NotFittedError
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from sklearn.model_selection import RandomizedSearchCV, GridSearchCV
import warnings

class ShapRFECV:
    """
    This class performs Backwards Recursive Feature Elimination, using SHAP features importance. At each round, for a given
     features set, starting from all available features, a model is optimized (e.g. using RandomSearchCV) and trained.
     At the end of each round, the n lowest SHAP feature importance features are removed and the model results are
     stored. The user can plot the performance of the model for each round, and select the optimal number of features
     and the features set.

    The functionality is similar to [sklearn.feature_selection.RFECV](https://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.RFECV.html),
     yet, it removes the lowest importance features based on SHAP features importance and optimizes the hyperparameters
     of the model at each round.

    We recommend using LightGBM model, because by default it handles missing values and categorical features. In case
     of other models, make sure to handle these issues for your dataset and consider impact it might have on features
     importance.

    Example:
    ```python
    from probatus.feature_elimination import ShapRFECV
    from sklearn.datasets import make_classification
    from sklearn.model_selection import train_test_split
    import numpy as np
    import pandas as pd
    import lightgbm

    feature_names = ['f1_categorical', 'f2_missing', 'f3_static', 'f4', 'f5', 'f6', 'f7', 'f8', 'f9', 'f10', 'f11', 'f12', 'f13', 'f14', 'f15']

    # Prepare two samples
    X, y = make_classification(n_samples=1000, n_features=15, random_state=0, n_redundant=10)
    X = pd.DataFrame(X, columns=feature_names)
    X['f1_categorical'] = X['f1_categorical'].apply(lambda x: str(np.round(x*10)))
    X['f2_missing'] = X['f2_missing'].apply(lambda x: x if np.random.rand()<0.8 else np.nan)
    X['f3_static'] = 0

    # Prepare model and parameter search space
    clf = lightgbm.LGBMClassifier(max_depth=5, class_weight='balanced')
    param_grid = {
        'n_estimators': [5, 7, 10],
        'num_leaves': [3, 5, 7],
    }

        # Run feature elimination
    shap_elimination = ShapRFECV(
        clf=clf, search_space=param_grid, search_schema='grid',
        step=0.2, cv=20, scoring='roc_auc', n_jobs=3, random_state=42)
    report = shap_elimination.fit_compute(X, y)

    # Make plots
    shap_elimination.plot('performance')
    shap_elimination.plot('parameter', param_names=['n_estimators', 'num_leaves'])

    # Get final features set
    final_features_set = shap_elimination.get_reduced_features_set(num_features=2)
    ```
    """

    def __init__(self, clf, search_space, search_schema='random', step=1, min_features_to_select=1,
                 random_state=None, **search_kwargs):
        """
        This method initializes the class:

        Args:
            clf (binary classifier): A model that will be optimized and trained at each round of features
             elimination. The recommended model is LightGBM, because it by default handles the missing values and
             categorical variables.

            search_space (dict of sklearn.ParamGrid): Parameter search space, which will be explored during the
             hyperparameter search. In case `grid` search_schema, it is passed to GridSearchCV as `param_grid`, in case
             of `random` search_schema, then this value is passed to RandomSearchCV as `param_distributions` parameter.

            search_schema (Optional, str): The hyperparameter search algorithm that should be used to optimize the model.
              It can be one of the following:

                - `random`: [RandomSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html)
                  which randomly selects hyperparameters from the prowided param_grid, and performs optimization using
                  Cross-Validation. It is recommended option, when you optimize a large number of hyperparameters.
                - `grid`: [GridSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html)
                  which searches through all permutations of hyperparameters values from param_grid, and performs
                  optimization using Cross-Validation. It is recommended option, for low number of hyperparameters.

            step (Optional, int or float): Number of lowest importance features removed each round. If it
             is an int, then each round such number of features is discarded. If float, such percentage of remaining
             features (rounded down) is removed each iteration. It is recommended to use float, since it is faster for a
             large number of features, and slows down and becomes more precise towards less features. Note: the last
             round may remove fewer features in order to reach min_features_to_select.

            min_features_to_select (Optional, unt): Minimum number of features to be kept. This is a stopping criterion
             of the feature elimination. By default the process stops when one feature is left.

            random_state (Optional, int): Random state set at each round of feature elimination. If it is None, the
             results will not be reproducible and in random search at each iteration a different hyperparameters might
             be tested. For reproducible results set it to integer.

            **search_kwargs: The keywords arguments passed to a given search schema, during initialization. Please refer
             to the parameters of a given search schema.
        """
        self.clf = clf

        if search_schema == 'random':
            self.search_class = RandomizedSearchCV
        elif search_schema == 'grid':
            self.search_class = GridSearchCV
        else:
            raise(ValueError('Unsupported search_schema, choose one of the following: "random", "grid".'))

        self.search_space = search_space

        self.search_kwargs = search_kwargs

        if (isinstance(step, int) or isinstance(step, float)) and \
                step > 0:
            self.step = step
        else:
            raise (ValueError(f"The current value of step = {step} is not allowed. "
                              f"It needs to be a positive integer or positive float."))

        if isinstance(min_features_to_select, int) and min_features_to_select>0:
            self.min_features_to_select=min_features_to_select
        else:
            raise (ValueError(f"The current value of min_features_to_select = {min_features_to_select} is not allowed. "
                              f"It needs to be a positive integer."))

        self.random_state = random_state

        self.report_df = pd.DataFrame([])
        self.fitted = False


    def _check_if_fitted(self):
        """
        Checks if object has been fitted. If not, NotFittedError is raised.
        """
        if self.fitted is False:
            raise(NotFittedError('The object has not been fitted. Please run fit() method first'))

    @staticmethod
    def _preprocess_data(X):
        """
        Does basic preprocessing of the data: Removal of static features, Warns which features have missing variables,
        and transform object dtype features to category type, such that LightGBM handles them by default.

        Args:
            X (pd.DataFrame): Provided dataset.

        Returns:
            (pd.DataFrame): Preprocessed dataset.
        """
        # Make sure that X is a pd.DataFrame
        X = assure_pandas_df(X)

        # Remove static features, those that have only one value for all samples
        static_features = [i for i in X.columns if len(X[i].unique()) == 1]
        if len(static_features)>0:
            warnings.warn(f'Removing static features {static_features}.')
            X = X.drop(columns=static_features)

        # Warn if missing
        columns_with_missing = [column for column in X.columns if X[column].isnull().values.any()]
        if len(columns_with_missing) > 0:
            warnings.warn(f'The following variables contain missing values {columns_with_missing}. Make sure to impute'
                          f'missing or apply a model that handles them automatically.')

        # Transform Categorical variables into category dtype
        indices_obj_dtype_features = [column[0] for column in enumerate(X.dtypes) if column[1] == 'O']
        obj_dtype_features = list(X.columns[indices_obj_dtype_features])

        # Set categorical features type to category
        if len(obj_dtype_features) > 0:
            warnings.warn(f'Changing dtype of {obj_dtype_features} from "object" to "category". Treating it as '
                          f'categorical variable. Make sure that the model handles categorical variables, or encode '
                          f'them first.')
            for obj_dtype_feature in obj_dtype_features:
                X[obj_dtype_feature] = X[obj_dtype_feature].astype('category')
        return X


    def _get_current_features_to_remove(self, shap_importance_df):
        """
        Implements the logic used to determine which features to remove. If step is a positive integer,
        at each round step lowest SHAP importance features are selected. If it is a float, such percentage
        of remaining features (rounded up) is removed each iteration. It is recommended to use float, since it is faster
        for a large set of features, and slows down and becomes more precise towards less features.

        Args:
            shap_importance_df (pd.DataFrame): DataFrame presenting SHAP importance of remaining features.

        Returns:
            (list): List of features to be removed at a given round.
        """

        # If the step is an int remove n features.
        if isinstance(self.step, int):
            num_features_to_remove = self._calculate_number_of_features_to_remove(
                current_num_of_features=shap_importance_df.shape[0],
                num_features_to_remove=self.step,
                min_num_features_to_keep=self.min_features_to_select
            )
        # If the step is a float remove n * number features that are left, rounded down
        elif isinstance(self.step, float):
            current_step = int(np.floor(shap_importance_df.shape[0] * self.step))
            # The step after rounding down should be at least 1
            if current_step < 1:
                current_step = 1

            num_features_to_remove = self._calculate_number_of_features_to_remove(
                current_num_of_features=shap_importance_df.shape[0],
                num_features_to_remove=current_step,
                min_num_features_to_keep=self.min_features_to_select
            )

        if num_features_to_remove == 0:
            return []
        else:

            return shap_importance_df.iloc[-num_features_to_remove:].index.tolist()


    @staticmethod
    def _calculate_number_of_features_to_remove(current_num_of_features, num_features_to_remove,
                                                min_num_features_to_keep):
        """
        Calculates the number of features to be removed, and makes sure that after removal at least
         min_num_features_to_keep are kept

         Args:
            current_num_of_features (int): Current number of features in the data.
            num_features_to_remove (int): Number of features to be removed at this stage.
            min_num_features_to_keep (int): Minimum number of features to be left after removal.

        Returns:
            (int) Number of features to be removed.
        """
        num_features_after_removal = current_num_of_features - num_features_to_remove
        if num_features_after_removal >= min_num_features_to_keep:
            num_to_remove = num_features_to_remove
        else:
            # take all available features minus number of them that should stay
            num_to_remove = current_num_of_features - min_num_features_to_keep
        return num_to_remove


    def _report_current_results(self, round_number, current_features_set, features_to_remove, search):
        """
        This function adds the results from a current iteration to the report.

        Args:
            round_number (int): Current number of the round.
            current_features_set (list of str): Current list of features.
            features_to_remove (list of str): List of features to be removed at the end of this iteration.
            search (sklearn.GridSearchCV or sklearn.RandomSearchCV): The fitted hyperparameter search object, containing
             results of the optimization.
        """
        current_results = {
            'num_features': len(current_features_set),
            'features_set': None,
            'eliminated_features':  None,
            'train_metric_mean': np.round(search.cv_results_['mean_train_score'][search.best_index_], 3),
            'train_metric_std': np.round(search.cv_results_['std_train_score'][search.best_index_], 3),
            'val_metric_mean': np.round(search.cv_results_['mean_test_score'][search.best_index_], 3),
            'val_metric_std': np.round(search.cv_results_['std_test_score'][search.best_index_], 3),
        }

        for param_name, param_value in search.best_params_.items():
            current_results[f'param_{param_name}'] = param_value

        current_row = pd.DataFrame(current_results, index=[round_number])
        current_row['features_set'] = [current_features_set]
        current_row['eliminated_features'] = [features_to_remove]

        self.report_df = pd.concat([self.report_df, current_row], axis=0)


    def fit(self, X, y):
        """
        Fits the object with the provided data. The algorithm starts with the entire dataset, and then sequentially
         eliminates features. At each step, it optimizes hyperparameters of the model, computes SHAP features importance
         and removes the lowest importance features.

        Args:
            X (pd.DataFrame): Provided dataset.
            y (pd.Series): Binary labels for X.
        """
        # Set seed for results reproducibility
        if self.random_state is not None:
            np.random.seed(self.random_state)

        self.X = self._preprocess_data(X)
        self.y = y

        remaining_features = current_features_set = self.X.columns.tolist()
        round_number = 0

        while len(current_features_set) > self.min_features_to_select:
            round_number += 1

            # Get current dataset info
            current_features_set = remaining_features
            current_X = self.X[current_features_set]

            # Optimize parameters
            # Set seed for results reproducibility
            if self.random_state is not None:
                np.random.seed(self.random_state)

            search = self.search_class(self.clf, self.search_space, refit=True,
                                       return_train_score=True, **self.search_kwargs)
            search = search.fit(current_X, y)

            # Compute SHAP values
            shap_values = shap_calc(search.best_estimator_, current_X, suppress_warnings=True)
            shap_importance_df = calculate_shap_importance(shap_values, remaining_features)

            # Get features to remove
            features_to_remove = self._get_current_features_to_remove(shap_importance_df)
            remaining_features = list(set(current_features_set) - set(features_to_remove))

            # Report results
            self._report_current_results(round_number=round_number, current_features_set=current_features_set,
                                         features_to_remove=features_to_remove, search=search)

            print(f'Round: {round_number}, Current number of features: {len(current_features_set)}, '
                  f'Current performance: Train {self.report_df.loc[round_number]["train_metric_mean"]} '
                  f'+/- {self.report_df.loc[round_number]["train_metric_std"]}, CV Validation '
                  f'{self.report_df.loc[round_number]["val_metric_mean"]} '
                  f'+/- {self.report_df.loc[round_number]["val_metric_std"]}. \n'
                  f'Num of features left: {len(remaining_features)}. '
                  f'Removed features at the end of the round: {features_to_remove}')
        self.fitted = True


    def compute(self):
        """
        Checks if fit() method has been run and computes the DataFrame with results of feature elimintation for each
         round.

        Returns:
            (pd.DataFrame): DataFrame with results of feature elimination for each round.
        """
        self._check_if_fitted()

        return self.report_df


    def fit_compute(self, X, y):
        """
        Fits the object and computes the report. The algorithm starts with the entire dataset, and then sequentially
         eliminates features. At each step, it optimizes hyperparameters of the model, computes SHAP features importance
         and removes the lowest importance features. At the end, the report containing results from each iteration is
         computed and returned to the user.

        Args:
            X (pd.DataFrame): Provided dataset.
            y (pd.Series): Binary labels for X.

        Returns:
            (pd.DataFrame): DataFrame containing results of feature elimination from each iteration.
        """

        self.fit(X, y)
        return self.compute()


    def get_reduced_features_set(self, num_features):
        """
        Gets the features set after the feature elimination process, for a given number of features.

        Args:
            num_features (int): Number of features in the reduced features set.

        Returns:
            (list of str): Reduced features set.
        """
        self._check_if_fitted()

        if num_features not in self.report_df.num_features.tolist():
            raise(ValueError(f'The provided number of features has not been achieved at any stage of the process. '
                             f'You can select one of the following: {self.report_df.num_features.tolist()}'))
        else:
            return self.report_df[self.report_df.num_features == num_features]['features_set'].values[0]


    def plot(self, plot_type='performance', param_names=None, show=True, **figure_kwargs):
        """
        Generates plots that allow to analyse the results.

        Args:
            plot_type (Optional, str): String indicating the plot type:

                - `performance`: Performance of the optimized model at each iteration. This plot allows to select the
                 optimal features set.
                - `parameter`: Plots the optimized hyperparameter's values at each iteration. This plot allows to
                 analyse stability of parameters for different features set. In case large variability of optimal
                 hyperparameters values is seen, consider reducing the search space.

            param_names (Optional, str, list of str): Name or names of parameters that will be plotted in case of
            `plot_type="parameter"`

            show (Optional, bool): If True, the plots are showed to the user, otherwise they are not shown.

            **figure_kwargs: Keyword arguments that are passed to the plt.figure, at its initialization.

        Returns:
            (plt.axis or list of plt.axis) Axis containing the target plot, or list of such axes.
        """
        x_ticks = list(reversed(self.report_df['num_features'].tolist()))

        if plot_type == 'performance':
            plt.figure(**figure_kwargs)

            plt.plot(self.report_df['num_features'], self.report_df['train_metric_mean'], label='Train Score')
            plt.fill_between(pd.to_numeric(self.report_df.num_features, errors='coerce'),
                             self.report_df['train_metric_mean'] - self.report_df['train_metric_std'],
                             self.report_df['train_metric_mean'] + self.report_df['train_metric_std'], alpha=.3)

            plt.plot(self.report_df['num_features'], self.report_df['val_metric_mean'], label='Validation Score')
            plt.fill_between(pd.to_numeric(self.report_df.num_features, errors='coerce'),
                             self.report_df['val_metric_mean'] - self.report_df['val_metric_std'],
                             self.report_df['val_metric_mean'] + self.report_df['val_metric_std'], alpha=.3)

            plt.xlabel('Number of features')
            plt.ylabel('Performance')
            plt.title('Backwards Feature Elimination using SHAP & CV')
            plt.legend(loc="lower left")
            ax = plt.gca()
            ax.invert_xaxis()
            ax.set_xticks(x_ticks)
            if show:
                plt.show()
            else:
                plt.close()

        elif plot_type == 'parameter':
            param_names = assure_list_of_strings(param_names, 'target_columns')
            ax = []
            for param_name in param_names:
                plt.figure(**figure_kwargs)

                plt.plot(self.report_df['num_features'], self.report_df[f'param_{param_name}'],
                         label=f'{param_name} optimized value')
                plt.xlabel('Number of features')
                plt.ylabel(f'Optimizal {param_name} value')
                plt.title(f'Optimization of {param_name} for different numbers of features')
                plt.legend(loc="lower left")
                current_ax = plt.gca()
                current_ax.invert_xaxis()
                current_ax.set_xticks(x_ticks)
                ax.append(current_ax)
                if show:
                    plt.show()
                else:
                    plt.close()
        else:
            raise(ValueError('Wrong value of plot_type. Select from "performance" or "parameter"'))

        if isinstance(ax, list) and len(ax) == 1:
            ax = ax[0]
        return ax




# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from .exceptions import NotFittedError, DimensionalityError, UnsupportedModelError
from .scoring import Scorer, get_scorers
from .arrayfuncs import assure_numpy_array, assure_pandas_df, check_1d, warn_if_missing, check_numeric_dtypes, \
    assure_column_names_consistency
from .shap_helpers import shap_calc, shap_to_df, calculate_shap_importance
from .warnings import ApproximationWarning
from ._utils import class_name_from_object, assure_list_of_strings, assure_list_values_allowed
from .tree import TreePathFinder
from .plots import plot_distributions_of_feature

__all__ = ['NotFittedError', 'DimensionalityError', 'UnsupportedModelError', 'Scorer', 'assure_numpy_array', 'assure_pandas_df', 'check_1d',
           'ApproximationWarning', 'class_name_from_object', 'get_scorers', 'assure_list_of_strings',
           'assure_list_values_allowed', 'warn_if_missing','TreePathFinder', 'check_numeric_dtypes',
           'plot_distributions_of_feature', 'assure_column_names_consistency', 'shap_calc', 'shap_to_df', 'calculate_shap_importance']



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from probatus.interpret import TreeDependencePlotter
from probatus.utils import assure_column_names_consistency, assure_pandas_df, shap_calc, assure_list_of_strings,\
    calculate_shap_importance, NotFittedError
from sklearn.metrics import roc_auc_score
import numpy as np
import shap
import matplotlib.pyplot as plt
import pandas as pd

class ShapModelInterpreter:
    """
    This class is a wrapper that allows to easily analyse model's features. It allows to plot SHAP feature importance,
     SHAP summary plot and SHAP dependence plots.

    Example:
    ```python
    from sklearn.datasets import make_classification
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.model_selection import train_test_split
    import numpy as np
    import pandas as pd

    feature_names = ['f1', 'f2', 'f3', 'f4']

    # Prepare two samples
    X, y = make_classification(n_samples=1000, n_features=4, random_state=0)
    X = pd.DataFrame(X, columns=feature_names)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

    # Prepare and fit model. Remember about class_weight="balanced" or an equivalent.
    clf = RandomForestClassifier(class_weight='balanced', n_estimators = 100, max_depth=2, random_state=0)
    clf.fit(X_train, y_train)

    # Train ShapModelAnalyser
    shap_interpreter = ShapModelInterpreter(clf)
    feature_importance = shap_interpreter.fit_compute(X_train, X_test, y_train, y_test)

    # Make plots
    shap_interpreter.plot('importance')
    shap_interpreter.plot('summary')
    shap_interpreter.plot('dependence', target_columns=['f1', 'f2'])
    shap_interpreter.plot('sample', samples_index=[521, 78])
    ```
    """


    def __init__(self, clf):
        """
        Initializes the class.

        Args:
            clf (binary classifier): Model fitted on X_train.
        """
        self.clf = clf
        self.fitted = False


    def _check_if_fitted(self):
        if self.fitted is False:
            raise(NotFittedError('The object has not been fitted. Please run fit() method first'))


    def fit(self, X_train, X_test, y_train, y_test, column_names=None, class_names=None, approximate=False,
            **shap_kwargs):
        """
        Fits the object and calculates the shap values for the provided datasets.

        Args:
            X_train (pd.DataFrame): Dataframe containing training data.
            X_test (pd.DataFrame): Dataframe containing test data.
            y_train (pd.Series): Series of binary labels for train data.
            y_test (pd.Series): Series of binary labels for test data.
            column_names (Optional, None, or list of str): List of feature names for the dataset. If None, then column
             names from the X_train dataframe are used.
            class_names (Optional, None, or list of str): List of class names e.g. ['neg', 'pos']. If none, the default
             ['Negative Class', 'Positive Class'] are used.
            approximate (boolean):, if True uses shap approximations - less accurate, but very fast.
            **shap_kwargs: keyword arguments passed to shap.TreeExplainer.
        """

        self.X_train = assure_pandas_df(X_train)
        self.X_test = assure_pandas_df(X_test)
        self.y_train = y_train
        self.y_test = y_test

        # Set class names
        self.class_names = class_names
        if self.class_names is None:
            self.class_names = ['Negative Class', 'Positive Class']

        # Set column names
        self.column_names = assure_column_names_consistency(column_names, self.X_train)

        # Calculate Metrics
        self.auc_train = roc_auc_score(self.y_train, self.clf.predict_proba(self.X_train)[:, 1])
        self.auc_test = roc_auc_score(self.y_test, self.clf.predict_proba(self.X_test)[:, 1])
        self.results_text = "Train AUC: {},\nTest AUC: {}.".format(
            np.round(self.auc_train, 3),
            np.round(self.auc_test, 3)
        )

        self.shap_values, self.explainer = shap_calc(self.clf, self.X_test, approximate=approximate,
                                                     return_explainer=True, data=self.X_train, **shap_kwargs)

        # Get expected_value from the explainer
        self.expected_value = self.explainer.expected_value
        # For sklearn models the expected values consists of two elements (negative_class and positive_class)
        if isinstance(self.expected_value, list) or isinstance(self.expected_value, np.ndarray):
            self.expected_value = self.expected_value[1]

        # Initialize tree dependence plotter
        self.tdp = TreeDependencePlotter(self.clf).fit(self.X_test, self.y_test, precalc_shap=self.shap_values)

        self.fitted = True


    def compute(self):
        """
        Computes the DataFrame, that presents the importance of each feature.

        Returns:
            (pd.DataFrame): Dataframe with SHAP feature importance.
        """
        self._check_if_fitted()

        # Compute SHAP importance
        self.importance_df = calculate_shap_importance(self.shap_values, self.column_names)
        return self.importance_df


    def fit_compute(self,  X_train, X_test, y_train, y_test, column_names=None, class_names=None, approximate=False,
                    **shap_kwargs):
        """
        Fits the object and calculates the shap values for the provided datasets.

        Args:
            X_train (pd.DataFrame): Dataframe containing training data.
            X_test (pd.DataFrame): Dataframe containing test data.
            y_train (pd.Series): Series of binary labels for train data.
            y_test (pd.Series): Series of binary labels for test data.
            column_names (Optional, None, or list of str): List of feature names for the dataset. If None, then column
             names from the X_train dataframe are used.
            class_names (Optional, None, or list of str): List of class names e.g. ['neg', 'pos']. If none, the default
             ['Negative Class', 'Positive Class'] are used.
            approximate (boolean):, if True uses shap approximations - less accurate, but very fast.
            **shap_kwargs: keyword arguments passed to shap.TreeExplainer.

        Returns:
            (pd.DataFrame): Dataframe with SHAP feature importance.
        """
        self.fit(X_train=X_train, X_test=X_test, y_train=y_train, y_test=y_test, column_names=column_names,
                 class_names=class_names, approximate=approximate, **shap_kwargs)
        return self.compute()


    def plot(self, plot_type, target_columns=None, samples_index=None, **plot_kwargs):
        """
        Plots the appropriate SHAP plot

        Args:
            plot_type (str): One of the following:

                - 'importance': Feature importance plot, SHAP bar summary plot
                - 'summary': SHAP Summary plot
                - 'dependence': Dependence plot for each feature
                - 'sample': Explanation of a given sample in the test data

            target_columns (Optional, None, str or list of str): List of features names, for which the plots should be
             generated. If None, all features will be plotted.

            samples_index (None, int, list or pd.Index): Index of samples to be explained if the `plot_type=sample`.

            **plot_kwargs: Keyword arguments passed to the plot method. For 'importance' and 'summary' plot_type, the
             kwargs are passed to shap.summary_plot, for 'dependence' plot_type, they are passed to
             probatus.interpret.TreeDependencePlotter.feature_plot method.

        """
        if target_columns is None:
            target_columns = self.column_names

        target_columns = assure_list_of_strings(target_columns, 'target_columns')
        target_columns_indices = [self.column_names.index(target_column) for target_column in target_columns]

        if plot_type in ['importance', 'summary']:
            # Get the target features
            target_X_test = self.X_test[target_columns]
            target_shap_values = self.shap_values[:, target_columns_indices]

            # Set summary plot settings
            if plot_type == 'importance':
                plot_type = 'bar'
                plot_title = 'SHAP Feature Importance'
            else:
                plot_type = 'dot'
                plot_title = 'SHAP Summary plot'

            shap.summary_plot(target_shap_values, target_X_test, plot_type=plot_type,
                              class_names=self.class_names, show=False, **plot_kwargs)
            ax = plt.gca()
            ax.set_title(plot_title)

            ax.annotate(self.results_text, (0, 0), (0, -50), fontsize=12, xycoords='axes fraction',
                        textcoords='offset points', va='top')
            plt.show()
        elif plot_type == 'dependence':
            ax = []
            for feature_name in target_columns:
                print()
                ax.append(
                    self.tdp.plot(feature=feature_name, figsize=(10, 7), target_names=self.class_names))
                plt.show()

        elif plot_type == 'sample':
            # Ensure the correct samples_index type
            if samples_index is None:
                raise(ValueError('For sample plot, you need to specify the samples_index be plotted plot'))
            elif isinstance(samples_index, int) or isinstance(samples_index, str):
                samples_index = [samples_index]
            elif not(isinstance(samples_index, list) or isinstance(samples_index, pd.Index)):
                raise(TypeError('sample_index must be one of the following: int, str, list or pd.Index'))

            ax = []
            for sample_index in samples_index:
                sample_loc = self.X_test.index.get_loc(sample_index)

                shap.plots._waterfall.waterfall_legacy(self.expected_value,
                                                       self.shap_values[sample_loc, :],
                                                       self.X_test.loc[sample_index],
                                                       show=False, **plot_kwargs)

                plot_title = f'SHAP Sample Explanation for index={sample_index}'
                current_ax = plt.gca()
                current_ax.set_title(plot_title)
                ax.append(current_ax)
                plt.show()
        else:
            raise ValueError("Wrong plot type, select from 'importance', 'summary', or 'dependence'")

        if isinstance(ax, list) and len(ax) == 1:
            ax = ax[0]
        return ax


# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import warnings

import numpy as np
import scipy.stats as stats

from ..utils import assure_numpy_array


def psi(d1, d2, verbose=False):
    """
    Calculates the Population Stability Index

    A simple statistical test that quantifies the similarity of two distributions. Commonly used in the banking / risk
    modeling industry. Only works on categorical data or bucketed numerical data. Distributions must be binned/bucketed
    before passing them to this function. Bin boundaries should be same for both distributions. Distributions must have
    same number of buckets. Note that the PSI varies with number of buckets chosen (typically 10-20 bins are used).
    Quantile bucketing is typically recommended.

    Args:
        d1 (np.ndarray or pd.core.series.Series) : first distribution ("expected").

        d2 (np.ndarray or pd.core.series.Series) : second distribution ("actual").

        verbose (bool)                           : print useful interpretation info to stdout (default False).

    Returns:
        (float, float) : measure of the similarity between d1 & d2. (range 0-inf, with 0 indicating identical
        distributions and > 0.25 indicating significantly different distributions); p_value for rejecting null
        hypothesis (that the two distributions are identical).
    """
    # Perform data checks
    d1 = assure_numpy_array(d1)
    d2 = assure_numpy_array(d2)

    if len(d1) < 10:
        warnings.warn("PSI is not well-behaved when using less than 10 bins.")
    if len(d1) > 20:
        warnings.warn("PSI is not well-behaved when using more than 20 bins.")
    if len(d1) != len(d2):
        raise ValueError("Distributions do not have the same number of bins.")

    # Number of bins/buckets
    b = len(d1)

    # Calculate the number of samples in each distribution
    n = d1.sum()
    m = d2.sum()

    # Calculate the ratio of samples in each bin
    expected_ratio = d1 / n
    actual_ratio = d2 / m

    # Necessary to avoid divide by zero and ln(0). Should have minor impact on PSI value.
    for i in range(0, b):
        if expected_ratio[i] == 0:
            expected_ratio[i] = 0.0001
            if verbose:
                print(
                    "PSI: Bucket {} has zero counts; may result in over-estimated (larger) PSI value. Decreasing \
                        the number of buckets may also help avoid buckets with zero counts.".format(i))

    # Calculate the PSI value
    psi_value = np.sum(
        (actual_ratio - expected_ratio) * np.log(actual_ratio / expected_ratio)
    )

    # Print the evaluation of statistical hypotheses
    if verbose:
        print("\nPSI =", psi_value)

        print("\nPSI: Critical values defined according to de facto industry standard:")
        if psi_value <= 0.1:
            print("PSI <= 0.10: No significant distribution change.")
        elif 0.1 < psi_value <= 0.25:
            print("PSI <= 0.25: Small distribution change; may require investigation.")
        elif psi_value > 0.25:
            print("PSI > 0.25: Significant distribution change; investigate.")

        # Calculate the critical values and
        alpha = [0.95, 0.99, 0.999]
        z_alpha = stats.norm.ppf(alpha)
        psi_critvals = ((1 / n) + (1 / m)) * (b - 1) + z_alpha * (
            (1 / n) + (1 / m)
        ) * np.sqrt(2 * (b - 1))
        print("\nPSI: Critical values defined according to Yurdakul (2018):")
        if psi_value > psi_critvals[2]:
            print("99.9% confident distributions have changed.")
        elif psi_value > psi_critvals[1]:
            print("99% confident distributions have changed.")
        elif psi_value > psi_critvals[0]:
            print("95% confident distributions have changed.")
        elif psi_value < psi_critvals[0]:
            print("No significant distribution change.")

    # Calculate p-value
    z = (psi_value / ((1 / n) + (1 / m)) - (b - 1)) / np.sqrt(2 * (b - 1))
    p_value = stats.norm.cdf(z)

    return psi_value, p_value



import numpy as np
import pytest

from probatus.binning import SimpleBucketer, QuantileBucketer, AgglomerativeBucketer, TreeBucketer
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted


@pytest.mark.filterwarnings("ignore:")
def test_deprecations():
    x = [1, 2, 1]
    bins = 3
    myBucketer = SimpleBucketer(bin_count=bins)
    myBucketer.fit(x)
    with pytest.deprecated_call():
        myBucketer.counts

    with pytest.deprecated_call():
        myBucketer.boundaries
    
    
def test_simple_bins():
    x = [1, 2, 1]
    bins = 3
    myBucketer = SimpleBucketer(bin_count=bins)
    with pytest.raises(NotFittedError):
       myBucketer.compute([1,2])
    
    myBucketer.fit(x)
    assert len(myBucketer.counts_) == bins
    assert np.array_equal(myBucketer.counts_, np.array([2, 0, 1]))
    assert len(myBucketer.boundaries_) == bins + 1
    np.testing.assert_array_almost_equal(myBucketer.boundaries_, np.array([1., 1.33333333, 1.66666667, 2.]))
    # test static method
    counts, boundaries = SimpleBucketer(bin_count=bins).simple_bins(x, bins)
    assert np.array_equal(myBucketer.counts_, counts)
    np.testing.assert_array_almost_equal(myBucketer.boundaries_, boundaries)
    assert repr(myBucketer).startswith('SimpleBucketer')


def test_quantile_bins():
    bins = 4
    random_state = np.random.RandomState(0)
    x = random_state.normal(0, 1, size=1000)
    myBucketer = QuantileBucketer(bin_count=bins)
    with pytest.raises(NotFittedError):
       myBucketer.compute([1,2])
    myBucketer.fit(x)
    assert len(myBucketer.counts_) == bins
    assert np.array_equal(myBucketer.counts_, np.array([250, 250, 250, 250]))
    assert len(myBucketer.boundaries_) == bins + 1
    np.testing.assert_array_almost_equal(myBucketer.boundaries_, np.array([-3.0, -0.7, -0.1, 0.6, 2.8]), decimal=1)
    # test static method
    counts, boundaries = QuantileBucketer(bin_count=bins).quantile_bins(x, bins)
    assert np.array_equal(myBucketer.counts_, counts)
    np.testing.assert_array_almost_equal(myBucketer.boundaries_, boundaries)
    # test inf edges
    counts, boundaries = QuantileBucketer(bin_count=bins).quantile_bins(x, bins, inf_edges=True)
    assert boundaries[0] == -np.inf
    assert boundaries[-1] == np.inf
    assert repr(myBucketer).startswith('QuantileBucketer')


def test_agglomerative_clustering_new():
    def log_function(x):
        return 1 / (1 + np.exp(-10 * x))

    x = [log_function(x) for x in np.arange(-1, 1, 0.01)]
    bins = 4
    myBucketer = AgglomerativeBucketer(bin_count=bins)
    with pytest.raises(NotFittedError):
       myBucketer.compute([1,2])
    myBucketer.fit(x)
    assert len(myBucketer.counts_) == bins
    assert np.array_equal(myBucketer.counts_, np.array([24, 16, 80, 80]))
    assert len(myBucketer.boundaries_) == bins + 1
    np.testing.assert_array_almost_equal(myBucketer.boundaries_, np.array([0, 0.11, 0.59, 0.88, 0.99]), decimal=2)
    # test static method
    counts, boundaries = AgglomerativeBucketer(bin_count=bins).agglomerative_clustering_binning(x, bins)
    assert np.array_equal(myBucketer.counts_, counts)
    np.testing.assert_array_almost_equal(myBucketer.boundaries_, boundaries)
    assert repr(myBucketer).startswith('AgglomerativeBucketer')


def test_compute():
    x = np.arange(10)
    bins = 5
    myBucketer = QuantileBucketer(bins)
    x_new = x
    with pytest.raises(NotFittedError):
        assert myBucketer.compute(x_new)
    myBucketer.fit(x)
    assert len(myBucketer.compute(x_new)) == bins
    np.testing.assert_array_equal(myBucketer.counts_, myBucketer.compute(x_new))
    np.testing.assert_array_equal(myBucketer.counts_, myBucketer.fit_compute(x_new))
    x_new = x + 100
    np.testing.assert_array_equal(np.array([0, 0, 0, 0, 0]), myBucketer.compute(x_new))
    x_new = x - 100
    np.testing.assert_array_equal(np.array([10, 0, 0, 0, 0]), myBucketer.compute(x_new))
    x_new = [1, 1, 1, 4, 4, 7]
    np.testing.assert_array_equal(np.array([3, 0, 2, 1, 0]), myBucketer.compute(x_new))


def test_quantile_with_unique_values():
    np.random.seed(42)
    dist_0_1 = np.random.uniform(size=20)
    dist_peak_at_0 = np.zeros(shape=20)

    skewed_dist = np.hstack((dist_0_1, dist_peak_at_0))
    actual_out = QuantileBucketer(10).quantile_bins(skewed_dist, 10)

    expected_out = (
        np.array([20, 4, 4, 4, 4, 4]),
        np.array([0., 0.01894458, 0.23632033, 0.42214475, 0.60977678, 0.67440958, 0.99940487])
    )

    assert (actual_out[0] == expected_out[0]).all()


def test_tree_bucketer():

    x = np.array([0. , 0.2, 0.4, 0.6, 0.8, 1. , 1.2, 1.4, 1.6, 1.8, 2. , 2.2, 2.4,
       2.6, 2.8, 3. , 3.2, 3.4, 3.6, 3.8, 4. , 4.2, 4.4, 4.6, 4.8, 5. ,
       5.2, 5.4, 5.6, 5.8, 6. , 6.2, 6.4, 6.6, 6.8, 7. , 7.2, 7.4, 7.6,
       7.8, 8. , 8.2, 8.4, 8.6, 8.8, 9. , 9.2, 9.4, 9.6, 9.8])

    y = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
       1, 0, 0, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1, 1,
       1, 1, 1, 1, 1, 1])

    myTreeBucketer = TreeBucketer(inf_edges=False,max_depth=3,min_samples_leaf=10,random_state=42)

    with pytest.raises(NotFittedError):
       myTreeBucketer.compute([1,2])

    myTreeBucketer.fit(x,y)

    assert all(myTreeBucketer.counts_==np.array([21, 15, 14]))
    assert myTreeBucketer.bin_count==3
    assert all(myTreeBucketer.boundaries_ - np.array([0.0, 4.1, 7.1, 9.8])<0.01)

    # If infinite edges is False, it must get the edges of the x array
    assert myTreeBucketer.boundaries_[0] == 0
    assert myTreeBucketer.boundaries_[-1] == 9.8


    myTreeBucketer = TreeBucketer(inf_edges=True,max_depth=3, min_samples_leaf=10, random_state=42)

    myTreeBucketer.fit(x, y)
    # check that the infinite edges is True, then edges must be infinite
    assert myTreeBucketer.boundaries_[0] == -np.inf
    assert myTreeBucketer.boundaries_[-1] == +np.inf


def test_tree_bucketer_dependence():
    x = np.arange(0, 10, 0.01)
    y = [1 if z < 0.5 else 0 for z in np.random.uniform(size=x.shape[0])]


    # Test number of leaves is always within the expected ranges
    myTreeBucketer = TreeBucketer(inf_edges=False, max_depth=3, min_samples_leaf=10, random_state=42).fit(x,y)
    assert myTreeBucketer.bin_count<=np.power(2,myTreeBucketer.tree.max_depth)

    # Test number of leaves is always within the expected ranges
    myTreeBucketer = TreeBucketer(inf_edges=False, max_depth=6, min_samples_leaf=1, random_state=42).fit(x, y)
    assert myTreeBucketer.bin_count <= np.power(2, myTreeBucketer.tree.max_depth)

    # Test that the counts per bin never drop below min_samples_leaf
    myTreeBucketer = TreeBucketer(inf_edges=False, max_depth=6, min_samples_leaf=100, random_state=42).fit(x, y)
    assert all(myTreeBucketer.counts_ >= myTreeBucketer.tree.min_samples_leaf)

    myTreeBucketer = TreeBucketer(inf_edges=False, max_depth=6, min_samples_leaf=200, random_state=42).fit(x, y)
    assert all(myTreeBucketer.counts_ >= myTreeBucketer.tree.min_samples_leaf)

    # Test that if the leaf is set to the number of entries,it raises an Error
    myTreeBucketer = TreeBucketer(inf_edges=False, max_depth=6, min_samples_leaf=x.shape[0], random_state=42)

    with pytest.raises(ValueError):
        assert myTreeBucketer.fit(x, y)

    # Test that if the leaf is set to the number of entries-1, it returns only one bin
    myTreeBucketer = (
        TreeBucketer(inf_edges=False, max_depth=6, min_samples_leaf=x.shape[0]-1, random_state=42) \
        .fit(x, y)
    )
    assert myTreeBucketer.bin_count ==1
    assert all(myTreeBucketer.counts_ >= myTreeBucketer.tree.min_samples_leaf)


